/*
 * NAME:
 *	netio - interface stdio to read_net and write_net
 *
 * SYNOPSIS:
 *	#include <netio.h>
 *
 *	int	net_setvbuf(NETIOB *p,
 *				char *buf, int	type, size_t  size);
 *	NETIOB *	net_getiob(int	fd, int force, int w);
 *	NETIOB *	net_fgetiobt(FILE *fp, int (*test)(int));
 *	NETIOB *	net_fgetiob(FILE *fp);
 *	int	net_filbuf(NETIOB *p);
 *	int	net_flsbuf(int	c, NETIOB *p);
 *	int	net_fflush(FILE *fp);
 *	int	net_fclose(FILE *fp);
 *	int	net_fprintf(FILE *fp, const char *fmt, ...);
 *	int	net_fgetc(FILE *fp);
 *	int	net_ungetc(int c, FILE *fp);
 *	int	net_fputc(int  c, FILE *fp);
 *	char *	net_fgets(char *buf, int  nbytes, FILE *fp);
 *	char *	net_fgetln(FILE *fp, size_t *n);
 *	int	net_read(int fd, char *buf, int nbytes);
 *	int	net_write(int fd, char *buf, int nbytes);
 *
 *
 * DESCRIPTION:
 *	These routines simply allow encryption hooks to be used with
 *	applications that use stdio.
 *
 *	'netio.h' contains macro definitions which override fprintf(),
 *	vfprintf(), fgets(), fgetc(), getc(), putc() and fputc().
 *
 *	Each of the net_*() functions, checks whether the FILE * is
 *	associated with an SSL connection and if not calls the normal
 *	stdio function, otherwise it implements the appropriate
 *	buffered I/O behaviour with net_filbuf() and net_flsbuf()
 *	calling read_net() and write_net() respectively.
 *	
 * SEE ALSO:
 *	rwnet.c
 *
 * BUGS:
 *	The implementation of net_vfprintf() is worth noting...
 *	We simply sprintf() into a big buffer, and if the result is
 *	too big we abort().  Ugly, but very portable. 
 *	 
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 *	
 *	This is a re-work of a stdio to socket implementation for DOS
 *	I did back in the late 1980s.
 */
/*
 *      @(#)Copyright (c) 1989,1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: netio.c,v 1.11 1997/05/15 15:49:05 sjg Exp $";
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <syslog.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <assert.h>

#ifdef USE_SSL
# include "ssl.h"
# include "sslfd.h"
#endif

#include "netiob.h"

int
net_vfprintf(fp, fmt, ap)
        FILE *fp;
        const char *fmt;
        va_list ap;
{
#ifdef NEED_NETIOB
	NETIOB *p = net_fgetiob(fp);
	static char buf[8192];
        size_t ret, n;
	
	if (!p)
		return vfprintf(fp, fmt, ap);

	/*
	 * This is the simplest way to do it
	 */
	p->_flag |= _IOWRT;
#ifdef HAVE_VSNPRINTF
	n = vsnprintf(buf, sizeof(buf) - 1, fmt, ap);
#else
	(void) vsprintf(buf, fmt, ap);
	n = strlen(buf);
#ifdef ASSERT
	ASSERT(n < sizeof(buf));
#else
	if (n >= sizeof(buf)) {
		syslog(LOG_ERR, "assertion failed %d < %d", n, sizeof(buf));
		abort();
	}
#endif
#endif
	for (ret = 0; ret < n; ret++)
		/*
		 * Interesting, using the macro here trashed the stack!
		 */
		if (net_putc(buf[ret], p) == EOF)
			return EOF;
	return ret;
#else
	return vfprintf(fp, fmt, ap);
#endif
}


int
#if __STDC__
net_fprintf(FILE *fp, const char *fmt, ...)
#else
net_fprintf(fp, fmt, va_alist)
        FILE *fp;
        char *fmt;
        va_dcl
#endif
{
        int ret;
        va_list ap;
 
#if __STDC__
        va_start(ap, fmt);
#else
        va_start(ap);
#endif
        ret = net_vfprintf(fp, fmt, ap);
        va_end(ap);
        return (ret);
}



int
net_fgetc(fp)
	FILE *fp;
{
#ifdef NEED_NETIOB
	NETIOB *p = net_fgetiob(fp);
	
	if (p)
		return net_getc(p);
#endif
	return getc(fp);
}


int
_net_ungetc(c, p)
	int c;
	NETIOB *p;
{
#ifndef NEED_NETIOB
	return ungetc(c, p);
#else
	if (c == EOF)
		return c;

	if (p->_base > 0) {
		if (p->_ptr >= p->_base) {
			*--p->_ptr = c;
			p->_cnt++;
			if (p->_ptr < p->_base)
				p->_ptr = p->_base;
		}
	}
	return c;
#endif
}

int
net_ungetc(c, fp)
	int c;
	FILE *fp;
{
#ifdef NEED_NETIOB
	NETIOB *p;

	if (c == EOF)
		return c;
	
	p = net_fgetiob(fp);
	
	if (p)
		return _net_ungetc(c, p);
#endif
	return ungetc(c, fp);
}



int
net_fputc(c, fp)
	int c;
	FILE *fp;
{
#ifdef NEED_NETIOB
	NETIOB *p = net_fgetiob(fp);

	if (p) {
		p->_flag |= _IOWRT;
		return net_putc(c, p);
	}
#endif
	return putc(c, fp);
}

char *
net_fgets(buf, nbytes, fp)
	char *buf;
	int nbytes;
	FILE *fp;
{
#ifdef NEED_NETIOB
	NETIOB *p = net_fgetiob(fp);
	char *s = buf;
	int c;
	
	if (buf == 0 || nbytes <= 0 || fp == 0)
		return 0;

	if (!p)
		return fgets(buf, nbytes, fp);
	
	nbytes--;
	while (nbytes) {
		c = net_getc(p);
		if (c == EOF) {
			if (s == buf)
				return 0;
			break;
		}
		*s++ = c;
		if (c == '\n')
			break;
	}
	*s = 0;
	return buf;
#else
	return fgets(buf, nbytes, fp);
#endif
}
