/*
 * NAME:
 *	ssl - routines useful for SSL servers and clients
 *
 * SYNOPSIS:
 *	SSL_CTX *sslfd_ctx(void)
 *	int ssl_setcon(int "fd", SSL *"con")
 *	SSL *ssl_getcon(int "fd")
 *	int set_sslcon(int "fd", SSL *"con")
 *	SSL *get_sslcon(int "fd")
 *	int ssl_pending(int "fd")
 *	int ssl_close(int "fd")
 *	void ssl_err(int "log_lev", char *"fmt")
 *	int ssl_get_pw(const char *"key_file", int "log_lev")
 *	int ssl_init(const char *"cert_file",
 *		const char *"key_file", const char *"cipher_list",
 *		int "verify_flag", int (*"verify_cb")(),
 *		int "log_lev", int "flags")
 *	int ssl_start1(int "s", int "log_lev", int "active",
 *		SSL *"con", int (*"cb")(), int "vf")
 *	int ssl_start_cb(int "s", int "log_lev", int "active",
 *		int (*"cb")(), int "vf")
 *	int ssl_start(int "s", int "log_lev", int "active")
 *	int ssl_clone_cb(int "s", int "old", int "log_lev",
 *		int "active", int (*"cb")(), int "vf")
 *	int ssl_clone(int "s", int "old", int "log_lev", int "active")
 *	int ssl_start2_cb(int "s", int "old", int "log_lev",
 *		int "active", int (*"cb")(), int "vf")
 *	int ssl_start2(int "s", int "log_lev", int "active")
 *	char *get_x509_name(char *"buf", int "bufsz")
 *	int ssl_setopt(const char *"optarg")
 *	
 * DESCRIPTION:
 *	These routines provide for client authentication using
 *	Eric Young's SSLeay library.
 *
 *	If no other definition of a certificate to use can be found,
 *	we will attempt to use _PATH_SSL_CERT (/etc/host.cert).
 *
 *	ssl_* vars set via ssl_setopt() now override the args to
 *	ssl_init() so they can be an applications defaults.
 *	
 * SEE ALSO:
 *	sslfd(3), ssld(3), sslc(3), ssl_rcmd(3), netio(3).
 *
 * NOTES:
 *	The current version of SSLeay can be obtained from:
 *	'ftp://ftp.psy.uq.oz.au/pub/Crypto/SSL/SSLeay-x.xx.tar.gz'
 *
 *	The SSLeay FAQ by Tim Hudson <tjh@mincom.oz.au> can be found
 *	at 'http://www.psy.uq.oz.au/~ftp/Crypto'
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *	@(#)Copyright (c) 1996 Simon J. Gerraty.
 *	
 *	This is free software.	It comes with NO WARRANTY.
 *	Permission to use, modify and distribute this source code 
 *	is granted subject to the following conditions.
 *	1/ that the above copyright notice and this notice 
 *	are preserved in all copies and that due credit be given 
 *	to the author.	
 *	2/ that any changes to this code are clearly commented 
 *	as such so that the author does not get blamed for bugs 
 *	other than his own.
 *	
 *	Please send copies of changes and bug-fixes to:
 *	sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: ssl.c,v 1.43 1998/11/12 04:01:10 sjg Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#ifndef R_OK
#include <sys/file.h>
#endif

#include "x509.h"
#include "ssl.h"
#include "crypto.h"

#define _EXTERN
#include "sslfd.h"

/* make sure we get cipher's we trust */
/* SSLv3 seems to have dropped the DES-*-MD5 ciphers so include the SHA ones */
#ifndef SSL_CIPHER_LIST
# ifdef SSLEAY8
/*
 * This is the same as SSL_DEFAULT_CIPHER_LIST but with EXP grades
 * removed.
 */
#   ifndef NO_RSA
#     define SSL_CIPHER_LIST "!EXP:!ADH:RC4+RSA:HIGH:MEDIUM:LOW:+SSLv2"
#   else
#     define SSL_CIPHER_LIST "!EXP:HIGH:MEDIUM:LOW:ADH+3DES:ADH+RC4:ADH+DES"
#   endif
# else
#   define SSL_CIPHER_LIST "RC4-SHA:RC4-MD5:IDEA-CBC-SHA:IDEA-CBC-MD5:DES-CBC3-SHA:DES-CBC3-MD5:DES-CBC-SHA:DES-CBC-MD5"
# endif
#endif
#ifndef SSL_VERIFY_CLIENT_ONCE
# define SSL_VERIFY_CLIENT_ONCE 0
#endif
#ifndef SSL_VERIFY_FLAG
# define SSL_VERIFY_FLAG SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE
#endif
#ifndef NUMFDS
# define NUMFDS 64
#endif

static SSL_CTX	*ssl_ctx = 0;
static SSL	*ssl_con;
char x509_name[256];
char *MyId = "sslfd";

/* these are set by ssl_setopt() */
static char *ssl_DHparam_file = 0;
static char *ssl_cert_file = 0;
static char *ssl_key_file = 0;
static char *defer_key_file = 0;
static char *ssl_cipher_list = 0;

static SSL *ssl_cons[NUMFDS];

#ifdef SSLEAY8
static SSL_METHOD *ssl_meth = 0;
/*
 * this cb is based on the one in s_server.c
 */
static RSA *
tmp_rsa_cb(con, export)
	SSL *con;
	int export;
{
	static RSA *tmp_rsa = 0;

	if (tmp_rsa == 0) {
#ifndef NO_RSA
# if SSLEAY_VERSION_NUMBER >= 0x0900
		tmp_rsa = RSA_generate_key(512,RSA_F4,NULL,NULL);
# else
		tmp_rsa = RSA_generate_key(512,RSA_F4,NULL);
# endif
#endif
	}
	return tmp_rsa;
}
#else
/*
 * these just keep the code below easier to read.
 */
# define SSL_CTX_use_PrivateKey_file(x, f, t) \
	SSL_use_RSAPrivateKey_file(ssl_con, (f), (t))
# define SSL_CTX_use_certificate_file(x, f, t) \
	SSL_use_certificate_file(ssl_con, (f), (t))
#endif

SSL_CTX	*
sslfd_ctx()
{
	return ssl_ctx;
}

int
ssl_setcon(fd, con)
	int fd;
	SSL *con;
{
	static int once = 1;	

	if (once) {
		memset((char *)ssl_cons, 0, sizeof(ssl_cons));
		once = 0;
	}
	if (fd >= 0 && fd < NUMFDS) {
		ssl_cons[fd] = con;
		return 1;
	}
	return 0;
}

SSL *
ssl_getcon(fd)
	int fd;
{
	static int once = 1;

	if (once) {
		ssl_setcon(-1, 0);		/* ensure initialized */
		once = 0;
	}
	if (fd >= 0 && fd < NUMFDS) {
		return ssl_cons[fd];
	}
	return 0;
}

/* old names */
int
set_sslcon(fd, con)
	int fd;
	SSL *con;
{
	return ssl_setcon(fd, con);
}

SSL *
get_sslcon(fd)
	int fd;
{
	return ssl_getcon(fd);
}

int
ssl_pending(fd)
	int fd;
{
	SSL *con;
	
	if (con = ssl_getcon(fd)) {
		return SSL_pending(con);
	}
	return 0;
}


int
ssl_close(fd)
	int fd;
{
	SSL *con;
	int i, n;

	if (con = ssl_cons[fd]) {
		SSL_shutdown(con);
		if (con != ssl_con) {		/* want to keep this one */
			for (i = n = 0; i < NUMFDS; i++) {
				if (ssl_cons[i] == con)
					n++;
			}
			if (n == 1) {
				SSL_free(con);	/* last reference */
			}
		}
		ssl_setcon(fd, 0);
	}
	return close(fd);
}

static long last_err;

void
ssl_err(log_lev, fmt)
	int log_lev;
	char *fmt;
{
	last_err = ERR_get_error();
	if (log_lev > 0) {
		syslog(log_lev, fmt, ERR_error_string(last_err, NULL));
	} else if (log_lev < 0) {
		fprintf(stderr, "%s: ", MyId);
		fprintf(stderr, fmt, ERR_error_string(last_err, NULL));
	}
}

/*
 * this is mainly here for telnet's use
 */
int
ssl_get_pw(key_file, log_lev)
	const char *key_file;
	int log_lev;
{
#ifdef SSLEAY8
	X509 *x509;
	ASN1_UTCTIME *utc;
#endif
	
	if (!key_file)
		key_file = defer_key_file;
	
	if (key_file && *key_file) {
		if (SSL_CTX_use_PrivateKey_file(ssl_ctx, 
						(char *) key_file,
						X509_FILETYPE_PEM)<=0) {
			ssl_err(log_lev, "PrivateKey_file error %s\n");
			return (-1);
		}
		if (key_file == defer_key_file) {
			free(defer_key_file);
			defer_key_file = 0;
		}
#ifdef SSLEAY8
		x509 = SSL_get_certificate(ssl_con);

		if (x509) {
			EVP_PKEY_copy_parameters(X509_get_pubkey(x509),
				SSL_get_privatekey(ssl_con));
		}
		if (!SSL_CTX_check_private_key(ssl_ctx)) {
			ssl_err(log_lev, "PrivateKey does not match cert\n");
			return (-1);
		}
		/*
		 * issue a warning if cert will expire soon...
		 */
		utc = X509_get_notAfter(x509);
		if (utcSecsCmp(utc, 0, 30*24*3600) < 0) {
			char *name = X509_NAME_ONELINE(X509_get_subject_name(x509));
			if (log_lev > 0) {
				syslog(log_lev,
				       "WARNING: expires: %s%.6s: cert=%s",
				       (utc->data[0] == '9') ? "19" : "20",
				       (char *) utc->data, name);
				       
			} else {
				fprintf(stderr,
					"%s: WARNING: expires: %s%.6s: cert=%s\n",
					MyId,
					(utc->data[0] == '9') ? "19" : "20",
					(char *) utc->data, name);
			}
			free(name);
		}
#endif
	}
	return 0;
}

int
ssl_init(cert_file, key_file, cipher_list, verify_flag, verify_cb, log_lev, flags)
	const char *cert_file, *key_file, *cipher_list;
	int verify_flag;
	int (*verify_cb)();
	int log_lev;
	int flags;
{
#ifndef _PATH_SSL_CERT
# define _PATH_SSL_CERT "/etc/host.cert"
#endif
	int sts = -1;
	char *err = 0;
	
	/*
	 * if the ssl_* vars are set, someone called ssl_setopts()
	 * so use them in prference, thus args to ssl_init() can be
	 * defaults
	 */
	if (ssl_cert_file)
		cert_file = ssl_cert_file;
	if (!cert_file)
		cert_file = getenv("SSL_CERT");
	if (!cert_file && access(_PATH_SSL_CERT, R_OK) == 0)
		cert_file = _PATH_SSL_CERT;

	if (ssl_key_file)
		key_file = ssl_key_file;
	if (!key_file)
		key_file = getenv("SSL_KEY");
	if (ssl_cipher_list)
		cipher_list = ssl_cipher_list;
	if (!cipher_list)
		cipher_list = getenv("SSL_CIPHER");
#ifdef SSL_CIPHER_LIST
	if (!cipher_list)
		cipher_list = SSL_CIPHER_LIST;
#endif
	if (!ssl_DHparam_file)
		ssl_DHparam_file = getenv("SSL_DHPARAM");
	if (ssl_verify_flag < 0) {
		if (verify_flag >= 0)
			ssl_verify_flag = verify_flag;
		else if (ssl_verify_flag < -1)
			ssl_verify_flag = SSL_VERIFY_FLAG|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
		else if (ssl_verify_flag < 0)
			ssl_verify_flag = SSL_VERIFY_FLAG;
	}
	if (defer_key_file) {
		free(defer_key_file);
		defer_key_file = 0;
	}
	if (log_lev > 0)
		syslog(LOG_DEBUG, "cert_file=%s,key_file=%s,cipher_list=%s",
		       cert_file ? cert_file : "",
		       key_file ? key_file : "",
		       cipher_list ? cipher_list : ""
		       );
	
	SSL_load_error_strings();
#ifdef SSLEAY8
	/*
	 * this seems to be the essential magic
	 */
	SSLeay_add_ssl_algorithms();
	if (!ssl_meth)
		ssl_meth = SSLv23_method();
	if ((ssl_ctx = SSL_CTX_new(ssl_meth)) == 0) {
		ssl_err(log_lev, "cannot create SSL context: %s\n");
		return (-1);
	}
	SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL);
	SSL_CTX_set_tmp_rsa_callback(ssl_ctx,tmp_rsa_cb);

	if (!SSL_CTX_set_default_verify_paths(ssl_ctx)) {
		ssl_err(log_lev, "cannot load certificates via SSL_CTX_set_default_verify_paths %s\n");
		return (-1);
	}
	SSL_CTX_set_verify(ssl_ctx, ssl_verify_flag, verify_cb);
	if (cipher_list && *cipher_list)
		SSL_CTX_set_cipher_list(ssl_ctx,(char *) cipher_list);
#else
	if ((ssl_ctx  = SSL_CTX_new()) == 0) {
		ssl_err(log_lev, "cannot create SSL context: %s\n");
		return (-1);
	}

	if (!X509_set_default_verify_paths(ssl_ctx->cert)) {
		ssl_err(log_lev, "cannot load certificates via X509_set_default_verify_paths %s\n");
		return (-1);
	}
#endif
	ssl_con = SSL_new(ssl_ctx);
	if (cert_file && *cert_file && access(cert_file, R_OK) == 0) {
		if (!(key_file && *key_file))
			key_file = cert_file;

		if (SSL_CTX_use_certificate_file(ssl_ctx,
						 (char *) cert_file,
						 X509_FILETYPE_PEM)<=0) {
			ssl_err(log_lev, "certificate_file error %s\n");
			return (-1);
		}
		if (key_file) {
			if (flags & 2)		/* defer key read */
				defer_key_file = strdup(key_file);
			else {
				if (ssl_get_pw(key_file, log_lev) < 0) {
					return (-1);
				}
			}
		}
	} else {
		if (log_lev > 0)
			syslog(LOG_DEBUG, "no cert_file");
	}
#ifndef SSLEAY8
	SSL_set_verify(ssl_con, ssl_verify_flag, verify_cb);
	if (cipher_list && *cipher_list)
		SSL_set_cipher_list(ssl_con,(char *) cipher_list);
#endif
	return (cert_file != 0);
}

/*
 * this does the real work for ssl_{start,clone}*()
 */
int
ssl_start1(s, log_lev, active, con, cb, vf)
	int s;
	int log_lev;
	int active;
	SSL	*con;
	int (*cb) ();
	int vf;
{
	int sts = -1;
	
	if (con == 0) {
		if (ssl_con == 0) {
			if (ssl_init(0, 0, 0, -1, cb, log_lev, 1) < 0)
				return -1;
		}
		if (ssl_ctx && (con = SSL_new(ssl_ctx))) {
			/*
			 * This just picks up details of our own
			 * cert etc...
			 */
			memcpy((char *)con, (char *)ssl_con, sizeof(SSL));
		}
	}	
	if (con == 0) {
		return -1;
	}
	SSL_set_fd(con,s);

	/*
	 * <sjg> we defer reading key until actually needed
	 */
	if (defer_key_file) {
		/*
		 * NOTE: ssl_get_pw() assumes con==ssl_con
		 */
		if (ssl_get_pw(defer_key_file, log_lev) < 0) {
			return (-1);
		}
	}
	/*
	 * <sjg> this is confusing, but the acceptor always provides a
	 * cert, where as it is optional for the connector.
	 * In cases  were the application only cares about client
	 * authentication, we can reverse the SSL roles to cut down
	 * the startup overheads.
	 * Mutual auth with 512bit keys takes up to 4s on a i486DX33
	 */
	if (cb) {
		SSL_set_verify(con, (vf < 0) ? ssl_verify_flag : vf, cb);
#ifdef SSLEAY8
		/*
		 * SSL_set_verify() does not do this.
		 */
		X509_STORE_set_verify_cb_func(SSL_CTX_get_cert_store(ssl_ctx),
					      cb);
#endif
	}
	if (active) {
		if ((sts = SSL_connect(con)) <= 0)
			ssl_err(log_lev, "SSL_connect error %s");
		
	} else {
#ifndef NO_DH
		static int once = 1;

		if (once) {
			once = 0;
			setDHparams(ssl_ctx, ssl_DHparam_file);
		}
#endif
		if ((sts = SSL_accept(con)) <= 0)
			ssl_err(log_lev, "SSL_accept error %s");
	}
	if (sts <= 0) {
#ifdef DEBUG
		char cbuf[512];
		STACK *sk;
		int i;
		
		switch (last_err & 0xfff) {
		default:
			break;
		case SSL_R_NO_SHARED_CIPHER:
			sk=SSL_get_ciphers(con);

			for (i = 0; i < sk_num(sk); i++) {
				SSL_CIPHER_description(
						       (SSL_CIPHER *)sk_value(sk,i),
						       cbuf,sizeof(cbuf));
				if (log_lev < 0) {
					if (i == 0)
						fprintf(stderr, "\nI support\n");
					fprintf(stderr, "%s", cbuf);
				} else {
					syslog(log_lev, "%s", cbuf);
				}
			}
			break;
		}
#endif
		return (-1);
	}
	ssl_setcon(s, con);
	return sts;
}

int
ssl_start_cb(s, log_lev, active, cb, vf)
	int s;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	int sts;
	
	if ((sts = ssl_start1(s, log_lev, active, 0, cb, vf)) > 0)
		ssl_fd = s;
	return sts;
}


int
ssl_start(s, log_lev, active)
	int s;
	int log_lev;
	int active;
{
	return ssl_start_cb(s, log_lev, active, 0, -1);
}


int
ssl_clone_cb(s, old, log_lev, active, cb, vf)
	int s;
	int old;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	SSL *con, *con2 = SSL_new(ssl_ctx);
	int sts = -1;
	
	if (con2) {
		if (old < 0)
			old = ssl_fd;	/* last primary session started */
		if (con = ssl_getcon(old)) {
			SSL_copy_session_id(con2, con);
			sts = ssl_start1(s, log_lev, active, con2, cb, vf);
		} else
			SSL_free(con2);
	}
	return sts;
}


int
ssl_clone(s, old, log_lev, active)
	int s;
	int old;
	int log_lev;
	int active;
{
	return ssl_clone_cb(s, old, log_lev, active, 0, -1);
}

/*
 * some old names
 */
int
ssl_start2_cb(s, old, log_lev, active, cb, vf)
	int s;
	int old;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	return ssl_clone_cb(s, old, log_lev, active, cb, vf);
}

int
ssl_start2(s, log_lev, active)
	int s;
	int log_lev;
	int active;
{
	return ssl_clone_cb(s, -1, log_lev, active, 0, -1);
}

char *
get_x509_field(buf, bufsz, field)
	char *buf;
	int bufsz;
	char *field;
{
	char *p, *e;
	int n = 0;

	if (buf == 0) {
		buf = x509_name;
		bufsz = sizeof(x509_name);
	}
	if (ssl_auth_cert) {
		if (p = strstr(ssl_auth_cert, field)) {
			p += strlen(field);
			if (e = strpbrk(p, "/\r\n")) {
				n = e - p;
				if (n >= bufsz)
					n = bufsz - 1;
			} else
				n = bufsz - 1;
			(void) strncpy(buf, p, n);
			buf[n] = '\0';
			return buf;
		}
	}
	return 0;
}

char *
get_x509_name(buf, bufsz)
	char *buf;
	int bufsz;
{
	return get_x509_field(buf, bufsz, "/CN=");
}

char *
get_x509_user(buf, bufsz)
	char *buf;
	int bufsz;
{
#ifndef X509_USER_FIELD
	return get_x509_name(buf, bufsz);
#else
	static char *userFields[] = {
		X509_USER_FIELD ,
		"/CN=",
		0
	};
	char **uf, *user = 0;
	
	if (ssl_auth_cert) {
		for (uf = userFields; uf; uf++) {
			if ((user = get_x509_field(buf, bufsz, *uf)) != 0 &&
			    *user != 0)
				return user;
		}
	}
	return 0;
#endif
}

#define OPTARG_VAL(str, cp, key) ((cp = strchr(str, '=')) && strncmp(str, key, cp - str) == 0 && ++cp)

int
ssl_setopt(optarg)
	const char *optarg;
{
	static int use_ssl = 1;
	char *cp;
	
	if (OPTARG_VAL(optarg, cp, "cert"))
		ssl_cert_file = cp;
	else if (OPTARG_VAL(optarg, cp, "cipher"))
		ssl_cipher_list = cp;
	else if (OPTARG_VAL(optarg, cp, "key"))
		ssl_key_file = cp;
	else if (OPTARG_VAL(optarg, cp, "verify"))
		ssl_verify_flag = atoi(cp);
	else if (OPTARG_VAL(optarg, cp, "depth"))
		ssl_verify_depth = atoi(cp);
	else if (OPTARG_VAL(optarg, cp, "auth"))
		_path_ssl_users = cp;
	else if (OPTARG_VAL(optarg, cp, "proxy"))
		ssl_proxy = cp;
	else if (strcmp(optarg, "ssl") == 0 || strcmp(optarg, "ssl_only") == 0)
		use_ssl |= 4;
	else if (strcmp(optarg, "secure") == 0) {
		use_ssl |= 2;
		ssl_secure_flag = 1;
	} else if (strcmp(optarg, "certsok") == 0)
		ssl_certsok_flag = 1;
#ifdef SSLEAY8
	else if (OPTARG_VAL(optarg, cp, "DHparam"))
		ssl_DHparam_file = cp;
	else if (strcmp(optarg, "v2") == 0)
		ssl_meth = SSLv2_method();
	else if (strcmp(optarg, "v3") == 0)
		ssl_meth = SSLv3_method();
	else if (strcmp(optarg, "v23") == 0)
		ssl_meth = SSLv23_method();
#endif
	return use_ssl;
}
